import numpy as np

from .mid_body_spot import MidBodySpot


class MidBodyTrack:
    """
    Mid-body candidate track
    """

    def __init__(self, track_id: int):
        self.track_id = track_id
        self.spots: dict[int, MidBodySpot] = {}
        self.length = 0

    def add_spot(self, spot: MidBodySpot) -> None:
        self.spots[spot.frame] = spot
        spot.track_id = self.track_id
        self.length += 1
        # Add children recursively
        if spot.child_spot is not None:
            self.add_spot(spot.child_spot)

    def get_expected_distance(
        self, expected_positions: dict[int, list[int]], max_distance: float
    ) -> float:
        """
        Compute the average distance between mid-body expected positions and current
        track positions.
        """
        distances = []
        for frame, position in expected_positions.items():
            if frame not in self.spots:
                continue
            distances.append(
                np.linalg.norm(np.array(self.spots[frame].position) - np.array(position))
            )
        # If there are no frames in common, for sure track is not the right one
        if len(distances) == 0:
            return np.inf
        mean_distance = np.mean(distances)
        if mean_distance > max_distance:
            return np.inf
        return mean_distance



import os
from skimage import io

from cut_detector._widget import mitosis_track_generation
from cut_detector.data.tools import get_data_path


def test_open_track_generation_widget():
    # Just try to open the widget
    mitosis_track_generation()


def test_track_generation_widget(make_napari_viewer_proxy):
    viewer = make_napari_viewer_proxy()

    # Add video
    video = io.imread(os.path.join(get_data_path("videos"), "example_video.tif"))
    viewer.add_image(video, name="example_video")

    # Open the widget
    widget = mitosis_track_generation()

    # Run process
    widget(
        viewer.layers[0],
        get_data_path("models"),
        get_data_path("mitoses"),
        get_data_path("tracks"),
    )



import os
from matplotlib import pyplot as plt
import numpy as np
import pickle
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from scipy.signal import find_peaks
from skimage.feature import graycomatrix, graycoprops

from ..utils.bridges_classification.impossible_detection import ImpossibleDetection
from ..utils.bridges_classification.template_type import TemplateType
from ..utils.hidden_markov_models import HiddenMarkovModel
from ..utils.image_tools import smart_cropping
from ..utils.mitosis_track import MitosisTrack


class MtCutDetectionFactory:
    """
    Class to detect first and second MT cut.

    Args:
        margin (int): Number of pixels on each side of the midbody.

        # Bridges classification parameters

        coeff_height_peak (float): Coeff for the height of the peak.
        circle_radius (int): Middle circle radius.
        circle_min_ratio (float): Min ratio distance on circle between two peaks.
        coeff_width_peak (float): Coeff for the width of the peak.
        diff_radius (float): Distance between the min/max and middle circle.
        window_size (float): Window size for the peak detection (% of the circle perimeter).
        min_peaks_by_group (int): Minimum number of peaks by window for the peak detection.
        overlap (int): Overlap between two windows for the peak detection.
        template_type (TemplateType): Template for SVC model.

        # Light spot detection parameters

        intensity_threshold_light_spot (int): Intensity threshold for the light spot detection.
        h_maxima_light_spot (int): h for the h_maxima function (light spot detection).
        center_tolerance_light_spot (int): Center tolerance to not count the light spots that are
        too close to the center.
        min_percentage_light_spot (float): Minimum percentage of frames with light spots to
        consider the mitosis as a light spot mitosis.
        crop_size_light_spot (int): Size of the crop for the light spot detection.
        length_light_spot (int): Length of the video to check around the mt cut for the light spot
        detection.
    """

    def __init__(
        self,
        margin=50,
        coeff_height_peak=1.071,
        circle_radius=11,
        circle_min_ratio=0.2025,
        coeff_width_peak=0.25358974,
        diff_radius=4.54,
        window_size=0.04836,
        min_peaks_by_group=4,
        overlap=4,
        intensity_threshold_light_spot=350,
        h_maxima_light_spot=130,
        center_tolerance_light_spot=5,
        min_percentage_light_spot=0.1,
        crop_size_light_spot=20,
        length_light_spot=10,
        template_type=TemplateType.ALL,
    ) -> None:
        self.margin = margin
        self.coeff_height_peak = coeff_height_peak
        self.circle_radius = circle_radius
        self.circle_min_ratio = circle_min_ratio
        self.coeff_width_peak = coeff_width_peak
        self.diff_radius = diff_radius
        self.window_size = window_size
        self.min_peaks_by_group = min_peaks_by_group
        self.overlap = overlap
        self.intensity_threshold_light_spot = intensity_threshold_light_spot
        self.h_maxima_light_spot = h_maxima_light_spot
        self.center_tolerance_light_spot = center_tolerance_light_spot
        self.min_percentage_light_spot = min_percentage_light_spot
        self.crop_size_light_spot = crop_size_light_spot
        self.length_light_spot = length_light_spot
        self.template_type = template_type

    @staticmethod
    def _is_bridges_classification_impossible(mitosis_track: MitosisTrack) -> bool:
        """
        Bridges classification is impossible if:
        - no midbody spots
        - more than 2 daughter tracks
        - nucleus is near border
        - no midbody spot after cytokinesis
        """

        if mitosis_track.is_near_border:
            mitosis_track.key_events_frame["first_mt_cut"] = ImpossibleDetection.NEAR_BORDER
            mitosis_track.key_events_frame["second_mt_cut"] = ImpossibleDetection.NEAR_BORDER
            return True

        if len(mitosis_track.daughter_track_ids) >= 2:
            mitosis_track.key_events_frame[
                "first_mt_cut"
            ] = ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            mitosis_track.key_events_frame[
                "second_mt_cut"
            ] = ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            return True

        if not mitosis_track.mid_body_spots:
            mitosis_track.key_events_frame[
                "first_mt_cut"
            ] = ImpossibleDetection.NO_MID_BODY_DETECTED
            mitosis_track.key_events_frame[
                "second_mt_cut"
            ] = ImpossibleDetection.NO_MID_BODY_DETECTED
            return True

        if (
            max(mitosis_track.mid_body_spots.keys())
            < mitosis_track.key_events_frame["cytokinesis"]
        ):
            mitosis_track.key_events_frame[
                "first_mt_cut"
            ] = ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            mitosis_track.key_events_frame[
                "second_mt_cut"
            ] = ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            return True

        return False

    def _get_circle_data(
        self,
        radius: int,
        image: np.ndarray,
    ) -> tuple[list[tuple[int]], list[float], float, np.ndarray, np.ndarray]:
        """
        Compute useful data for a circle of a given radius around the mid body spot:

        - circle_positions: list of the coordinates of the pixels on the circle
        - intensities: list of the intensities of the pixels on the circle
        - mean_intensity: mean intensity of the pixels on the circle
        - peaks: list of the indexes of the peaks of the intensities on the circle
        - norm_peaks: list of the normalized indexes of the peaks of the intensities on the circle
        """
        # Get image shape
        size_x = image.shape[1]
        size_y = image.shape[0]

        # Get the mid-body spot coordinates, middle of the image
        mid_body_position = (size_x // 2, size_y // 2)

        # Get associated perimeter, angles and coordinates
        perimeter = round(2 * np.pi * radius)
        angles = np.linspace(0, 2 * np.pi, perimeter)
        circle_positions = []
        for angle in angles:
            pos_x = round(mid_body_position[1] + radius * np.cos(angle))
            pos_y = round(mid_body_position[0] + radius * np.sin(angle))
            if (pos_x, pos_y) not in circle_positions:
                circle_positions.append((pos_x, pos_y))

        intensities = []
        for position in circle_positions:
            # Get the mean intensity around the mid spot
            total, inc = 0, 0
            for k in range(-1, 2):
                for j in range(-1, 2):
                    pos_x = position[0] + k
                    pos_y = position[1] + j
                    if 0 <= pos_x < size_x and 0 <= pos_y < size_y:
                        total += image[pos_x, pos_y]
                        inc += 1
            mean = total / inc
            intensities.append(mean)

        # Detect the peaks of the intensities on the circle
        mean_intensity = np.mean(intensities)

        # Prevent the case where the peak is on the border so we detect 2 peaks instead of 1
        concatenated_intensities = intensities + intensities

        # Insert lower intensity at the beginning and at the end of the list
        concatenated_intensities.insert(0, concatenated_intensities[0] - 1)
        concatenated_intensities.append(concatenated_intensities[-1] - 1)

        peaks, _ = find_peaks(
            concatenated_intensities,
            height=mean_intensity * self.coeff_height_peak,
            distance=len(intensities) * self.circle_min_ratio,
            width=(None, round(len(intensities) * self.coeff_width_peak)),
        )

        peaks = [i - 1 for i in peaks]
        peaks = [p % len(circle_positions) for p in peaks]
        peaks = [p for p in peaks if peaks.count(p) >= 2]
        peaks = list(set(peaks))

        norm_peaks = [peak / len(circle_positions) for peak in peaks]

        return circle_positions, intensities, mean_intensity, peaks, norm_peaks

    def _get_peaks(
        self,
        all_positions: list[list[tuple[int]]],
        all_intensities: list[list[float]],
        all_peaks: list[np.ndarray],
        all_norm_peaks: list[np.ndarray],
    ) -> tuple[list[tuple[int]], list[float]]:
        """
        Post-processing of peaks detection.

        Returns kept peaks (circle index, peak position) and their intensities.
        A typical output would be:
            final_peaks = [(0, 30), (0, 62)]
            peaks_intensity = [440, 416]
        Peak at position 30 on circle 0 was kept, with an intensity of 440.
        Peak at position 62 on circle 0 was kept, with an intensity of 416.
        """
        # Create the windows
        windows_starts = np.linspace(
            0, 1 - self.window_size, num=round(self.overlap / self.window_size)
        )
        windows_ends = np.linspace(self.window_size, 1, num=round(self.overlap / self.window_size))
        windows = list(zip(windows_starts, windows_ends))

        # Group the peaks by windows
        group_peaks = (
            []
        )  # list (window) of list of tuple (index of the circle, position of the peak)
        for window in windows:
            group = []
            for i, (peaks, norm_peaks) in enumerate(zip(all_peaks, all_norm_peaks)):
                for peak, norm_peak in zip(peaks, norm_peaks):
                    if window[0] <= norm_peak < window[1]:
                        group.append((i, peak))
            group_peaks.append(group)

        # Remove groups that have less than min_peaks_by_group peaks
        new_group_peaks = []
        for group_peak in group_peaks:
            if len(group_peak) >= self.min_peaks_by_group and group_peak not in new_group_peaks:
                new_group_peaks.append(group_peak)

        # If there is more than 1 peak, make sure there is no duplicate
        if len(new_group_peaks) >= 2:
            # Sort by maximum number of peaks, or higher mean intensity
            def get_length_intensity_key(item):
                return (len(item), np.mean([all_intensities[p[0]][p[1]] for p in item]))

            new_group_peaks.sort(key=get_length_intensity_key, reverse=True)

            # One peak must appear in one group only
            # Remove the peaks that are already seen in previous groups
            for i in range(len(new_group_peaks) - 1):
                for j in range(i + 1, len(new_group_peaks)):
                    new_group_peaks[j] = [
                        p for p in new_group_peaks[j] if p not in new_group_peaks[i]
                    ]

            # Remove groups that have less than min_peaks_by_group peaks
            new_group_peaks = [g for g in new_group_peaks if len(g) >= self.min_peaks_by_group]

            # Compute the mean position of the peaks in each group
            position_groups = [
                sum(pos / len(all_positions[circle]) for circle, pos in group) / len(group)
                for group in new_group_peaks
            ]

            # Remove the groups that are too close to each other, i.e. distance < circle_min_ratio
            i = 0
            while i < len(position_groups) - 1:
                j = i + 1
                while j < len(position_groups):
                    first_position = (position_groups[i] - position_groups[j]) % 1
                    second_position = (position_groups[j] - position_groups[i]) % 1
                    if min(first_position, second_position) < self.circle_min_ratio:
                        if len(new_group_peaks[i]) > len(new_group_peaks[j]):
                            new_group_peaks.remove(new_group_peaks[j])
                            position_groups.remove(position_groups[j])
                            j -= 1
                        else:
                            new_group_peaks.remove(new_group_peaks[i])
                            position_groups.remove(position_groups[i])
                            i -= 1
                    j += 1
                i += 1

            # Sort again by maximum number of peaks and mean intensity of the peaks in the group
            new_group_peaks.sort(key=get_length_intensity_key, reverse=True)

            # Should not be possible to have more than 2 MT, then keep 2 best ones
            if len(new_group_peaks) > 2:
                new_group_peaks = new_group_peaks[:2]

        # Compute peaks_intensity with the mean intensity of the peaks
        peaks_intensity = [
            sum(all_intensities[circle][pos] for circle, pos in group) / len(group)
            for group in new_group_peaks
        ]

        # final_peaks for plot
        final_peaks = [p[0] for p in new_group_peaks]

        return final_peaks, peaks_intensity

    @staticmethod
    def _get_haralick_features(
        list_intensity: list[float],
        properties=None,  # avoid list as default value
    ) -> np.ndarray:
        """
        Compute a few Haralick features from a list of intensities.
        """
        if properties is None:
            properties = ["contrast", "dissimilarity", "homogeneity", "energy", "correlation"]

        # Convert list_intensity to 2D array and int
        list_intensity_2D = np.array(list_intensity).astype(int).reshape(1, -1)

        # Compute all haralick features
        haralick_features = graycomatrix(
            list_intensity_2D,
            distances=[1],
            angles=[0],
            levels=max(list_intensity_2D[0]) + 1,
            symmetric=True,
            normed=True,
        )
        return np.array(
            [graycoprops(haralick_features, property)[0][0] for property in properties]
        )

    def _plot_circles(
        self,
        filename,
        image,
        final_peaks: list[tuple[int]],
        all_positions: list[list[tuple[int]]],
        mean_intensity: list[float],
        peaks_intensity: list[float],
        list_intensity: list[float],
    ) -> None:
        """
        Plot the image with the circle and the peaks.
        """

        # Plot the intensities of the circle around the mid body spot
        plt.subplot(2, 1, 1)
        plt.title("Number of peaks: " + str(len(final_peaks)))

        plt.plot(range(len(list_intensity)), list_intensity)
        plt.axhline(y=mean_intensity[0] * self.coeff_height_peak, color="red")
        plt.plot(
            [p[1] for p in final_peaks], peaks_intensity, marker="o", markersize=4, color="red"
        )

        # plot the image without the bridge line
        plt.subplot(2, 2, 3)
        plt.imshow(image)

        # plot the image with the bridge line
        plt.subplot(2, 2, 4)
        plt.imshow(image)

        if len(final_peaks) > 0:
            coord_peaks = [all_positions[p[0]][p[1]] for p in final_peaks]
            if len(final_peaks) == 2:
                plt.plot(
                    [coord_peaks[0][1], coord_peaks[1][1]],
                    [coord_peaks[0][0], coord_peaks[1][0]],
                    color="red",
                )
            elif len(final_peaks) == 1:
                plt.plot(
                    coord_peaks[0][1], coord_peaks[0][0], marker="o", markersize=4, color="red"
                )
        for i in range(len(all_positions)):
            x_cercle = [all_positions[i][k][0] for k in range(len(all_positions[i]))]
            y_cercle = [all_positions[i][k][1] for k in range(len(all_positions[i]))]
            plt.plot(y_cercle, x_cercle, color="green")

        plt.title(filename)
        plt.show()

    def get_bridge_template(
        self,
        image: np.ndarray,
        plot_enabled: bool,
        filename=None,
    ) -> np.ndarray:
        """
        Return bridge image embedding according to template provided in parameters.
        """

        # Create the coordinates list of the circle around the mid body spot
        # NB: first circle in list is the middle circle
        circle_radius, diff_radius = self.circle_radius, self.diff_radius
        list_radius = [
            circle_radius,
            circle_radius + 0.25 * diff_radius,
            circle_radius - 0.25 * diff_radius,
            circle_radius + 0.5 * diff_radius,
            circle_radius - 0.5 * diff_radius,
            circle_radius + 0.75 * diff_radius,
            circle_radius - 0.75 * diff_radius,
            circle_radius + diff_radius,
            circle_radius - diff_radius,
        ]

        # Get useful data for each circle
        all_positions, all_intensities, all_mean_intensity = [], [], []
        all_peaks, all_norm_peaks = [], []
        for radius in list_radius:
            positions, intensities, mean_intensity, peaks, norm_peaks = self._get_circle_data(
                radius, image
            )
            all_positions.append(positions)
            all_intensities.append(intensities)
            all_mean_intensity.append(mean_intensity)
            all_peaks.append(peaks)
            all_norm_peaks.append(norm_peaks)

        # Get peaks data
        final_peaks, peaks_intensity = self._get_peaks(
            all_positions, all_intensities, all_peaks, all_norm_peaks
        )

        # Get useful Haralick features from middle center circle
        haralick_features = self._get_haralick_features(all_intensities[0])

        # Plot if enabled
        if plot_enabled:
            self._plot_circles(
                filename,
                image,
                final_peaks,
                all_positions,
                all_mean_intensity,
                peaks_intensity,
                all_intensities[0],
            )

        # Get the template of the bridge
        if self.template_type == TemplateType.ALL_WITHOUT_HARALICK:
            return np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.var(all_intensities[0]),  # 2 : variance of intensity on the circle
                    np.mean(all_intensities[0]),  # 3 : mean intensity on the circle
                ]
            )

        if self.template_type == TemplateType.NB_PEAKS:
            return np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                ]
            )

        if self.template_type == TemplateType.PEAKS_AND_INTENSITY:
            return np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                ]
            )

        if self.template_type == TemplateType.HARALICK:
            return haralick_features

        if self.template_type == TemplateType.ALL:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features))

        # Ablation study on All
        if self.template_type == TemplateType.ALL_ABLATION_1:
            template = np.array(
                [
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features))

        if self.template_type == TemplateType.All_Ablation2:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(all_intensities[0]),  # 1 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features))

        if self.template_type == TemplateType.All_Ablation3:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                ]
            )
            return np.concatenate((template, haralick_features))

        if self.template_type == TemplateType.All_Ablation4:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features[1:]))

        if self.template_type == TemplateType.All_Ablation5:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features[:1], haralick_features[2:]))

        if self.template_type == TemplateType.All_Ablation6:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features[:2], haralick_features[3:]))

        if self.template_type == TemplateType.All_Ablation7:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features[:3], haralick_features[4:]))

        if self.template_type == TemplateType.All_Ablation8:
            template = np.array(
                [
                    len(final_peaks),  # 0 : number of peaks
                    np.mean(peaks_intensity)
                    if len(final_peaks) > 0
                    else np.mean(all_intensities[0]),  # 1 : intensity of the peaks
                    np.mean(all_intensities[0]),  # 2 : mean intensity on the circle
                ]
            )
            return np.concatenate((template, haralick_features[:4], haralick_features[5:]))

        raise ValueError("Unknown template type")

    def get_bridge_class(
        self,
        img_bridge: np.ndarray,
        scaler: StandardScaler,
        clf: SVC,
        plot_enabled=True,
    ) -> int:
        # Make sure crop has expected size
        assert img_bridge.shape[0] == self.margin * 2 and img_bridge.shape[1] == self.margin * 2

        # Get template and reshape to 2D data
        template = self.get_bridge_template(img_bridge, plot_enabled=plot_enabled).reshape(1, -1)

        # Scale template
        scaled_template = scaler.transform(template)

        # Predict the class
        frame_class = clf.predict(scaled_template)
        if frame_class[0] == "A":
            return 0
        if frame_class[0] == "B":
            return 1
        return 2

    @staticmethod
    def apply_hmm(hmm_parameters, sequence):
        """
        Correct the sequence of classes using HMM.
        """
        # Define observation sequence
        obs_seq = np.asarray(sequence, dtype=np.int32)

        # Define HMM model & run inference
        model = HiddenMarkovModel(hmm_parameters["A"], hmm_parameters["B"], hmm_parameters["pi"])
        states_seq, _ = model.viterbi_inference(obs_seq)

        return states_seq

    def update_mt_cut_detection(
        self,
        mitosis_track: MitosisTrack,
        video: np.ndarray,
        scaler_path: str,
        model_path: str,
        hmm_bridges_parameters_file: str,
    ) -> None:
        """
        Update micro-tubules cut detection using bridges classification.
        """
        classification_impossible = self._is_bridges_classification_impossible(mitosis_track)

        if classification_impossible:
            return

        # Perform classification...
        ordered_mb_frames = sorted(mitosis_track.mid_body_spots.keys())
        first_mb_frame = ordered_mb_frames[0]
        last_mb_frame = ordered_mb_frames[-1]
        first_frame = max(first_mb_frame, mitosis_track.key_events_frame["cytokinesis"] - 2)  # -2?

        # Load the classifier and scaler
        with open(model_path, "rb") as f:
            classifier: SVC = pickle.load(f)
        # Load the scaler
        with open(scaler_path, "rb") as f:
            scaler: StandardScaler = pickle.load(f)

        list_class_bridges = []
        # Iterate over frames and get the class of the bridge
        for frame in range(first_frame, last_mb_frame + 1):
            min_x = mitosis_track.position.min_x
            min_y = mitosis_track.position.min_y

            # Get midbody coordinates
            mb_coords = mitosis_track.mid_body_spots[frame].position
            x_pos, y_pos = min_x + mb_coords[0], min_y + mb_coords[1]

            # Extract frame image and crop around the midbody Sir-tubulin
            frame_image = video[frame, :, :, :].squeeze().transpose(2, 0, 1)  # C, H, W
            crop = smart_cropping(frame_image, self.margin, x_pos, y_pos, pad=True)[0, ...]  # H, W

            # Get the class of the bridge
            bridge_class = self.get_bridge_class(crop, scaler, classifier, plot_enabled=False)
            list_class_bridges.append(bridge_class)

        # Make sure cytokinesis bridge is detected as A (no MT cut)
        relative_cytokinesis_frame = mitosis_track.key_events_frame["cytokinesis"] - first_frame
        if relative_cytokinesis_frame < 0 or list_class_bridges[relative_cytokinesis_frame] != 0:
            mitosis_track.key_events_frame[
                "first_mt_cut"
            ] = ImpossibleDetection.MT_CUT_AT_CYTOKINESIS
            mitosis_track.key_events_frame[
                "second_mt_cut"
            ] = ImpossibleDetection.MT_CUT_AT_CYTOKINESIS
            return

        # Read HMM parameters
        if not os.path.exists(hmm_bridges_parameters_file):
            raise FileNotFoundError(f"File {hmm_bridges_parameters_file} not found")
        hmm_parameters = np.load(hmm_bridges_parameters_file)

        # Correct the sequence with HMM
        seq_after_hmm = self.apply_hmm(hmm_parameters, list_class_bridges)

        # Get index of first element > 0 in sequence (first MT cut)
        first_mt_cut_frame_rel = next((i for i, x in enumerate(seq_after_hmm) if x > 0), -1)

        # Ignore if no MT cut detected
        if first_mt_cut_frame_rel == -1:
            mitosis_track.key_events_frame["first_mt_cut"] = ImpossibleDetection.NO_CUT_DETECTED
            mitosis_track.key_events_frame["second_mt_cut"] = ImpossibleDetection.NO_CUT_DETECTED
            return

        first_mt_cut_frame_abs = first_frame + first_mt_cut_frame_rel
        if mitosis_track.light_spot_detected(
            video,
            first_mt_cut_frame_abs,
            self.length_light_spot,
            self.crop_size_light_spot,
            self.h_maxima_light_spot,
            self.intensity_threshold_light_spot,
            self.center_tolerance_light_spot,
            self.min_percentage_light_spot,
        ):
            mitosis_track.key_events_frame["first_mt_cut"] = ImpossibleDetection.LIGHT_SPOT
            mitosis_track.key_events_frame["second_mt_cut"] = ImpossibleDetection.LIGHT_SPOT
            return

        # Update mitosis track accordingly
        mitosis_track.key_events_frame["first_mt_cut"] = first_mt_cut_frame_abs

        # Get index of first element > 1 in sequence (second MT cut)
        second_mt_cut_frame_rel = next((i for i, x in enumerate(seq_after_hmm) if x > 1), -1)

        # get the frame of the second MT cut
        if second_mt_cut_frame_rel == -1:
            mitosis_track.key_events_frame["second_mt_cut"] = ImpossibleDetection.NO_CUT_DETECTED
            return

        second_mt_cut_frame_abs = first_frame + second_mt_cut_frame_rel

        # Update mitosis track accordingly
        mitosis_track.key_events_frame["second_mt_cut"] = second_mt_cut_frame_abs



# Cellpose parameters
FLOW_THRESHOLD = 0.0
CELLPROB_THRESHOLD = 0.0
AUGMENT = True

# Tracking parameters (ratio of average spot size)
LINKING_MAX_DISTANCE_RATIO = 1
GAP_CLOSING_MAX_DISTANCE_RATIO = 1 / 2
MAX_FRAME_GAP = 3

# Metaphase links
FRAMES_AROUND_METAPHASE = 10  # look for metaphase candidate spots in this range

INTERPHASE_INDEX = 0  # interphase index for CNN classification model
METAPHASE_INDEX = 1  # metaphase index for CNN classification model

# Mid-body tracking
CYTOKINESIS_DURATION = 20  # number of frames to look for mid-body in between cells

# Minimum distance to border to consider is_near_border = true
MINIMUM_DISTANCE_TO_BORDER = 20



import os
import urllib.request

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def get_model_path(model_name: str) -> None:
    """
    Returns absolute path to model file.
    """
    sub_folder_to_create = None
    if model_name == "segmentation_model":
        path_end = "segmentation_model"
    elif model_name == "metaphase_model":
        path_end = "metaphase_cnn/metaphase_cnn.pt"
        sub_folder_to_create = "metaphase_cnn"
    elif model_name == "hmm_metaphase_parameters":
        path_end = "hmm_metaphase_parameters.npz"
    elif model_name == "hmm_bridges_parameters":
        path_end = "hmm_bridges_parameters.npz"
    elif model_name == "svc_scaler":
        path_end = "svc_bridges/scaler.pkl"
        sub_folder_to_create = "svc_bridges"
    elif model_name == "svc_model":
        path_end = "svc_bridges/model.pkl"
        sub_folder_to_create = "svc_bridges"
    else:
        raise ValueError(f"Unknown model name: {model_name}")

    local_path = os.path.join(CURRENT_DIR, path_end)

    if not os.path.exists(local_path):
        print(f"Downloading model {model_name}...")
        if sub_folder_to_create is not None:
            os.makedirs(os.path.join(CURRENT_DIR, sub_folder_to_create), exist_ok=True)
        urllib.request.urlretrieve(
            f"https://raw.githubusercontent.com//15bonte/cut-detector-models/main/models/{path_end}",
            local_path,
        )

    return local_path



import numpy as np
import albumentations as A

from cnn_framework.utils.data_sets.abstract_data_set import AbstractDataSet
from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.preprocessing import normalize_array


class MetaphaseCnnDataSet(AbstractDataSet):
    """
    Custom class to avoid loading images from folder.
    """

    def __init__(self, data, *args, **kwargs):
        # Not pythonic, but needed as super init calls generate_raw_images
        self.data = data
        super().__init__(*args, **kwargs)

    def set_transforms(self):
        height, width = self.params.input_dimensions.to_tuple(False)
        self.transforms = A.Compose(
            [
                A.PadIfNeeded(
                    min_height=height,
                    min_width=width,
                    border_mode=0,
                    value=0,
                    p=1,
                ),
                A.CenterCrop(height=height, width=width, p=1),
            ]
        )

    def generate_raw_images(self, filename):
        idx = int(filename.split(".")[0])
        nucleus_image = normalize_array(self.data[idx], None)  # C, H, W
        nucleus_image = np.moveaxis(nucleus_image, 0, -1)  # H, W, C
        return DatasetOutput(
            input=nucleus_image, target_array=np.asarray([0, 1, 0])
        )



import os
import pickle
from typing import Optional
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import ttest_1samp

from ..utils.mitosis_track import MitosisTrack
from ..constants.bridges import TIME_RESOLUTION
from ..utils.bridges_classification.impossible_detection import (
    ImpossibleDetection,
)

ALPHA = 0.05
MAX_FRAME = np.inf  # 48*6 <-> 48h * 6 frames/hour


def box_plot_cut_differences(
    cut_differences: list[int], show: bool, save_dir: Optional[str]
) -> None:
    """
    Plot box plot of cut differences
    """
    if len(cut_differences) == 0:
        return

    plt.figure()
    plt.boxplot(cut_differences)
    plt.title("Box plot of cut differences")
    plt.ylabel("Cut difference (min)")

    if show:
        plt.show()

    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, "first_mt_cut_differences.png"))


def plot_cut_distributions(
    first_cut_times: list[int],
    first_cut_times_gt: list[int],
    p_value: Optional[float],
    show: bool,
    save_dir: Optional[str],
) -> None:
    """
    Plot distribution of cut differences
    """

    plt.figure()

    for cut_times, name in zip(
        [first_cut_times, first_cut_times_gt], ["Predicted", "Ground truth"]
    ):
        if len(cut_times) == 0:  # ignore empty gt list
            continue
        curve_values = []
        nb_values = len(cut_times)
        number_of_cut = 0
        for i in range(max(cut_times) + 1):
            count = cut_times.count(i)
            number_of_cut += count
            curve_values.append(number_of_cut / nb_values)
        plt.plot(curve_values, label=name)

    # Quartiles and median
    predicted_median = np.median(first_cut_times)
    gt_median = np.median(first_cut_times_gt) if len(first_cut_times_gt) else None
    predicted_q1 = np.percentile(first_cut_times, 25)
    predicted_q3 = np.percentile(first_cut_times, 75)
    plt.axvline(
        x=predicted_q1,
        color="r",
        linestyle="--",
        linewidth=0.5,
        label=f"Detected Q1: {predicted_q1}",
    )
    plt.axvline(
        x=predicted_median,
        color="r",
        linestyle="-",
        linewidth=0.5,
        label=f"Detected median: {predicted_median}"
        + (f"vs {gt_median} Ground truth median" if gt_median else ""),
    )
    plt.axvline(
        x=predicted_q3,
        color="r",
        linestyle="--",
        linewidth=0.5,
        label=f"Detected Q3: {predicted_q3}",
    )

    # add t test result to legend
    if p_value is not None:
        plt.legend(loc="lower right", title=f"t test p-value: {p_value:.2f}")
    else:
        plt.legend(loc="lower right")

    plt.title(
        f"Proportion of MT cuts over time - {len(first_cut_times)} predicted mitoses"
        + (f"- {len(first_cut_times_gt)} ground truth mitoses" if len(first_cut_times_gt) else "")
    )
    plt.xlabel("Time (min)")
    plt.ylabel("Proportion of first cut")

    if show:
        plt.show()

    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, "first_mt_cut_distributions.png"))


def perform_t_test(cut_differences: list[int]) -> float:
    """
    Compute t-test of the differences, which is supposed to be 0.
    """
    if len(cut_differences) == 0:
        return None

    t_result = ttest_1samp(cut_differences, 0)
    p_value = t_result.pvalue

    # Output the results
    print(f"P-value: {p_value}")

    # Check if the result is statistically significant (e.g., using a significance level of 0.05)
    if p_value < ALPHA:
        print("Reject the null hypothesis: There is a statistically significant difference.")
    else:
        print(
            "Fail to reject the null hypothesis: There is no statistically significant difference."
        )

    return p_value


def print_weird_mitoses(selected_tracks: list[MitosisTrack], min_acceptable_frame=13) -> None:
    """
    Print useful information about weird mitoses.
    """
    ordered_tracks = [
        track
        for track in selected_tracks
        if track.key_events_frame["first_mt_cut"] - track.key_events_frame["cytokinesis"]
        <= min_acceptable_frame
    ]
    ordered_tracks.sort(
        key=lambda x: x.key_events_frame["first_mt_cut"] - x.key_events_frame["cytokinesis"]
    )

    print("Weird mitoses (early cut):")
    for mitosis_track in ordered_tracks:
        print("")
        print(
            f"Track: {mitosis_track.id}_{mitosis_track.mother_track_id}_to_{','.join(str(daughter) for daughter in mitosis_track.daughter_track_ids)}"
        )
        print(f"Key events frame: {mitosis_track.key_events_frame}")
        print(
            "video frame cut",
            mitosis_track.key_events_frame["first_mt_cut"] - mitosis_track.min_frame + 1,
        )
        if mitosis_track.gt_key_events_frame is not None:
            print(f"GT key events frame: {mitosis_track.gt_key_events_frame}")
            print(
                "video frame cut",
                mitosis_track.gt_key_events_frame["first_mt_cut"] - mitosis_track.min_frame + 1,
            )


def print_analysis_summary(
    mitosis_tracks: list[MitosisTrack],
    mitosis_results_summary: dict[int, int],
    first_cut_times_gt: list[int],
    gt_mitosis_results_summary: dict[int, int],
) -> None:
    print(f"\n Among the {len(mitosis_tracks)} mitoses detected, there are:")
    for cut_id, count in mitosis_results_summary.items():
        print(f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}")

    if len(first_cut_times_gt) == 0:
        return
    print(f"\n Among the {len(first_cut_times_gt)} mitoses annotated, there are:")
    for cut_id, count in gt_mitosis_results_summary.items():
        print(f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}")


def save_csv_results(mitosis_tracks: list[MitosisTrack], save_dir: Optional[str] = None) -> None:
    if save_dir is None:
        return

    csv_path = os.path.join(save_dir, "results.csv")
    # Create CSV results file if it does not exist
    if not os.path.exists(csv_path):
        with open(csv_path, "w") as f:
            f.write(
                "id;mother track;daughter track(s);metaphase frame;cytokinesis frame;first MT cut frame;second MT cut frame;first MT cut time;second MT cut time\n"
            )
        f.close()

    # Store useful results in global results file
    with open(csv_path, "a") as f:
        for mitosis_track in mitosis_tracks:
            f.write(f"{mitosis_track.id};")
            f.write(f"{mitosis_track.mother_track_id};")
            daughter_ids = ",".join([str(d) for d in mitosis_track.daughter_track_ids])
            f.write(f"{daughter_ids};")
            f.write(f"{mitosis_track.key_events_frame['metaphase']};")
            cytokinesis_frame = mitosis_track.key_events_frame["cytokinesis"]
            f.write(f"{cytokinesis_frame};")
            first_cut_frame = mitosis_track.key_events_frame["first_mt_cut"]
            f.write(f"{first_cut_frame};")
            second_cut_frame = mitosis_track.key_events_frame["second_mt_cut"]
            f.write(f"{second_cut_frame};")
            first_cut_time = (
                (first_cut_frame - cytokinesis_frame) * TIME_RESOLUTION
                if first_cut_frame >= 0
                else ""
            )
            f.write(f"{first_cut_time};")
            second_cut_time = (
                (second_cut_frame - cytokinesis_frame) * TIME_RESOLUTION
                if second_cut_frame >= 0
                else ""
            )
            f.write(f"{second_cut_time};\n")
    f.close()


def perform_results_saving(
    exported_mitoses_dir: str,
    show: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
) -> None:
    # Create save_dir if specified and it does not exist
    if save_dir is not None and not os.path.exists(save_dir):
        os.makedirs(save_dir)

    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track = pickle.load(f)

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Define lists and dictionaries to store results
    first_cut_times: list[int] = []  # time in absolute time
    first_cut_times_gt: list[int] = []  # time in absolute time
    selected_tracks: list[MitosisTrack] = []  # kept tracks
    cut_differences: list[int] = []
    mitosis_results_summary: dict[int, int] = {}
    gt_mitosis_results_summary: dict[int, int] = {}
    # Get first MT cut time
    for mitosis_track in mitosis_tracks:
        # Get first cut frame and start of cytokinesis frame
        cyto_frame = mitosis_track.key_events_frame["cytokinesis"]
        cut_frame = mitosis_track.key_events_frame["first_mt_cut"]

        if cut_frame < 0 or cut_frame > MAX_FRAME:
            # No cut detected, for some reason
            cut_time = None
        else:
            # At least one cut was actually detected
            assert cut_frame >= cyto_frame  # should not be possible
            # Add the difference time to the list
            cut_time = (cut_frame - cyto_frame) * TIME_RESOLUTION
            first_cut_times.append(cut_time)
            selected_tracks.append(mitosis_track)

        # Update summary
        cut_id = min(cut_frame, 0)  # 0 if normal, negative otherwise

        if cut_id not in mitosis_results_summary:
            mitosis_results_summary[cut_id] = 0
        mitosis_results_summary[cut_id] += 1

        # Ignore if information is not there
        if (
            mitosis_track.gt_key_events_frame is None
            or "first_mt_cut" not in mitosis_track.gt_key_events_frame
        ):
            continue

        if mitosis_track.gt_key_events_frame["first_mt_cut"] > MAX_FRAME:
            # Ignore mitoses that are too late
            continue

        cut_time_gt = (
            mitosis_track.gt_key_events_frame["first_mt_cut"]
            - mitosis_track.gt_key_events_frame["cytokinesis"]
        ) * TIME_RESOLUTION
        first_cut_times_gt.append(cut_time_gt)

        if cut_time is not None:
            cut_differences.append(cut_time - cut_time_gt)

        # Update ground truth summary
        if cut_id not in gt_mitosis_results_summary:
            gt_mitosis_results_summary[cut_id] = 0
        gt_mitosis_results_summary[cut_id] += 1

    # Display few details on mitoses with very early cut
    if verbose:
        print_weird_mitoses(selected_tracks)

    # Statistical test
    p_value = perform_t_test(cut_differences)

    # Print summaries
    print_analysis_summary(
        mitosis_tracks,
        mitosis_results_summary,
        first_cut_times_gt,
        gt_mitosis_results_summary,
    )

    # Save results in as csv file
    save_csv_results(mitosis_tracks, save_dir)

    # Plot results
    box_plot_cut_differences(cut_differences, show, save_dir)
    plot_cut_distributions(first_cut_times, first_cut_times_gt, p_value, show, save_dir)



import numpy as np


def re_organize_channels(image: np.ndarray) -> np.ndarray:
    """
    Expect a 4 dimensions image.
    Re-organize channels to get TXYC order.
    """
    if image.ndim != 4:
        raise ValueError("Expect a 4 dimensions image.")

    # Get dimension index of smallest dimension, i.e. channels
    channels_dimension_index = np.argmin(image.shape)
    channels_dimension = image.shape[channels_dimension_index]
    if channels_dimension != 3:
        raise ValueError(
            "Expect 3 channels: SiR-tubulin/MKLP1/Phase contrast, in that order."
        )
    # Put channels at the back
    image = np.moveaxis(image, channels_dimension_index, 3)

    # Get dimension index of second smallest dimension, i.e. time
    second_dimension_index = np.argsort(image.shape)[1]
    # Put time at the front
    image = np.moveaxis(image, second_dimension_index, 0)

    return image


def cell_counter_frame_to_video_frame(
    cell_counter_frame: int, nb_channels=4
) -> int:
    """
    Cell counter index starts at 1, just like Fiji.

    To count frame, it just concatenates all channels.
    For example, with 4 channels, frames 1, 2, 3 and 4 will be frame 1,
    frames 5, 6, 7 and 8 will be frame 2, etc.
    """
    return (cell_counter_frame - 1) // nb_channels



from cut_detector._widget import results_saving
from cut_detector.data.tools import get_data_path


def test_open_results_saving_widget():
    # Just try to open the widget
    results_saving()


def test_results_saving_widget():
    # Open the widget
    widget = results_saving()

    # Run process
    widget(get_data_path("mitoses"), get_data_path("results"))


if __name__ == "__main__":
    test_open_results_saving_widget()
    test_results_saving_widget()



from __future__ import annotations

from typing import Optional
import numpy as np


class MidBodySpot:
    """
    Mid-body candidate spot
    """

    def __init__(
        self,
        frame: int,
        position: list[int],
        intensity: Optional[float] = None,
        sir_intensity: Optional[float] = None,
        area: Optional[float] = None,
        circularity: Optional[float] = None,
    ):
        self.frame = frame
        self.position = position  # (x, y)
        self.intensity = intensity
        self.sir_intensity = sir_intensity
        self.area = area
        self.circularity = circularity

        self.parent_spot: Optional[MidBodySpot] = None
        self.child_spot: Optional[MidBodySpot] = None

        self.track_id: Optional[int] = None

    def distance_to(self, other_spot: MidBodySpot) -> float:
        return np.sqrt(
            (self.position[0] - other_spot.position[0]) ** 2
            + (self.position[1] - other_spot.position[1]) ** 2
        )



# from https://github.com/TreB1eN/HiddenMarkovModel_Pytorch
# cf https://en.wikipedia.org/wiki/Baum%E2%80%93Welch_algorithm#Multiple_sequences for clean formula
# and http://www.cs.cmu.edu/~roni/11661/2017_fall_assignments/shen_tutorial.pdf
# and https://courses.grainger.illinois.edu/ECE417/fa2021/lectures/lec16.pdf
# and http://users.isr.ist.utl.pt/~wurmd/Livros/school/Bishop%20-%20Pattern%20Recognition%20And%20Machine%20Learning%20-%20Springer%20%202006.pdf
# or for alternative http://bozeman.genome.washington.edu/compbio/mbt599_2006/hmm_scaling_revised.pdf

import torch

from cnn_framework.utils.display_tools import display_progress


class HiddenMarkovModel(object):
    """
    Hidden Markov self Class
    Parameters:
    -----------

    - N: Number of states.
    - T: numpy.array Transition matrix of size N by N
         stores probability from state i to state j.
    - E: numpy.array Emission matrix of size N by T (number of observations)
         stores the probability of observing  O_j  from state  S_i.
    - T0: numpy.array Initial state probabilities of size N.
    """

    def __init__(self, A, B, pi, epsilon=0.001, max_step=10):
        # Max number of iteration
        self.max_step = max_step
        # convergence criteria
        self.epsilon = epsilon
        # Number of possible states
        self.N = A.shape[0]
        # Number of possible observations
        self.M = B.shape[0]
        # Emission probability
        self.B = torch.tensor(B)
        # Transition matrix
        self.A = torch.tensor(A)
        # Initial state vector
        self.pi = torch.tensor(pi)
        # Used for later display
        self.prob_state_1 = []

    def belief_propagation(self, scores):
        return scores.view(-1, 1) + torch.log(self.A)

    def viterbi_inference(self, x):  # x: observing sequence
        """
        Be careful, all comments below are according to the paper and ignore the log operation
        """
        T = len(x)
        shape = [T, self.N]

        # Init_viterbi_variables
        psi = torch.zeros(shape, dtype=torch.float64)
        delta = torch.zeros_like(psi)
        q_star = torch.zeros([shape[0]], dtype=torch.int64)

        # log probability of emission sequence
        obs_prob_full = torch.log(self.B[x])  # b_j(O_t) for all j for all t , shape (T, N)

        # initialize with state starting log-priors
        delta[0] = torch.log(self.pi) + obs_prob_full[0]  # pi_i * b_i(O_1) , shape (T)

        for step, obs_prob in enumerate(
            obs_prob_full[1:]
        ):  # note that first observation is ignored
            belief = self.belief_propagation(delta[step, :])  # delta_(step)(i) * a_ij for all j
            # the inferred state by maximizing global function
            psi[step + 1] = torch.argmax(belief, 0)  # arg_max(delta_(step)(i) * a_ij for all j)
            # and update state and score matrices
            delta[step + 1] = (
                torch.max(belief, 0)[0] + obs_prob
            )  # max(delta_(step)(i) * a_ij for all j) * b_j(O_(step+1)), +1 since first observation is ignored

        # infer most likely last state
        q_star[T - 1] = torch.argmax(delta[T - 1, :], 0)

        for step in range(T - 1, 0, -1):
            # for every timestep retrieve inferred state
            state = q_star[step]
            state_prob = psi[step][state]
            q_star[step - 1] = state_prob

        return q_star, torch.exp(delta)  # turn scores back to probabilities

    def _forward(self, obs_prob_seq):
        """
        obs_prob_seq: b_i(O_t) for all i for all N, shape (T, N)
        """
        T = len(obs_prob_seq)  # number of time stamps
        alpha = torch.zeros([T, self.N], dtype=torch.float64)
        scale_factors = torch.zeros([T], dtype=torch.float64)  # C in the paper
        # initialize with state starting priors
        alpha_0 = self.pi * obs_prob_seq[0]  # pi_i * b_i(O_0) for all i , shape (N)
        # scaling factor c_0 at t=0
        scale_factors[0] = 1.0 / alpha_0.sum()
        # scaled belief at t=0
        alpha[0] = scale_factors[0] * alpha_0
        # propagate belief
        for step, obs_prob in enumerate(obs_prob_seq[1:]):
            # previous state probability
            prev_prob = alpha[step].unsqueeze(0)  # alpha_step(i)
            # transition prior
            prior_prob = torch.matmul(prev_prob, self.A)  # sum(alpha_step(i) * a_ij for all i)
            # forward belief propagation
            forward_score = (
                prior_prob * obs_prob
            )  # sum(alpha_step(i) * a_ij for all i) * b_j(O_(step+1))
            forward_prob = torch.squeeze(forward_score)  # alpha_(step+1)(i)
            # scaling factor
            scale_factors[step + 1] = 1 / forward_prob.sum()
            # Update forward matrix
            alpha[step + 1] = scale_factors[step + 1] * forward_prob

        return alpha, scale_factors

    def _backward(self, obs_prob_seq_rev, scale_factors, beta_normalization):
        T = len(obs_prob_seq_rev)  # number of time stamps
        beta = torch.zeros([T, self.N], dtype=torch.float64)
        # initialize with state ending priors
        beta[0] = scale_factors[T - 1] * torch.ones(
            [self.N], dtype=torch.float64
        )  # initialize to scale_factors[T - 1]
        # propagate belief
        for step, obs_prob in enumerate(obs_prob_seq_rev[:-1]):
            # next state probability
            next_prob = beta[step, :].unsqueeze(1)  # beta_step(j)
            # observation emission probabilities
            obs_prob_d = torch.diag(obs_prob)
            # transition prior
            prior_prob = torch.matmul(self.A, obs_prob_d)  # sum(b_j(O_(step+1)) a_ij for all i)
            # backward belief propagation
            backward_prob = torch.matmul(
                prior_prob, next_prob
            ).squeeze()  # sum(b_j(O_(step+1)) a_ij for all i) * beta_step(j)
            # Update backward matrix
            scale_factor = (
                scale_factors[T - 2 - step]
                if beta_normalization == "original"
                else 1 / backward_prob.sum()
            )
            beta[step + 1] = (
                scale_factor * backward_prob  # from forward pass
            )  # beta_(step+1)(j) (should be -1 but flip in the end)
        beta = torch.flip(beta, [0, 1])
        return beta

    def forward_backward(self, obs_prob_seq, beta_normalization):
        """
        runs forward backward algorithm on observation sequence
        Arguments
        ---------
        - obs_prob_seq : matrix of size T by N, where T is number of timesteps and
            N is the number of states
        Returns
        -------
        - forward : matrix of size T by N representing
            the forward probability of each state at each time step
        - backward : matrix of size T by N representing
            the backward probability of each state at each time step
        - posterior : matrix of size T by N representing
            the posterior probability of each state at each time step
        """
        alpha, scale_factors = self._forward(obs_prob_seq)
        obs_prob_seq_rev = torch.flip(obs_prob_seq, [0, 1])
        beta = self._backward(obs_prob_seq_rev, scale_factors, beta_normalization)

        return alpha, beta

    def re_estimate_transition(self, observed_sequences, alphas, betas):
        summed_xis, gammas, updated_gammas = [], [], []

        for idx, (observed_seq, alpha, beta) in enumerate(zip(observed_sequences, alphas, betas)):
            T = len(observed_seq)  # number of time stamps
            xi = torch.zeros([T - 1, self.N, self.N], dtype=torch.float64)
            for t in range(T - 1):
                tmp_0 = torch.matmul(
                    alpha[t].unsqueeze(0), self.A
                )  # alpha_t(i) * a_ij for all i for all j, shape (1, N)
                tmp_1 = tmp_0 * self.B[observed_seq[t + 1]].unsqueeze(
                    0
                )  # alpha_t(i) * a_ij * b_j(O_(t+1)) for all i for all j, shape (1, N)
                denom = torch.matmul(
                    tmp_1, beta[t + 1].unsqueeze(1)
                ).squeeze()  # alpha_t(i) * a_ij * b_j(O_(t+1)) * beta_(t+1)(j) for all i for all j, scalar

                xi_t = torch.zeros([self.N, self.N], dtype=torch.float64)

                for i in range(self.N):
                    # alpha_t(i) * a_ij * b_j(O_(t+1)) * beta_(t+1)(j)
                    numer = alpha[t, i] * self.A[i, :] * self.B[observed_seq[t + 1]] * beta[t + 1]
                    xi_t[i] = numer / denom

                xi[t] = xi_t

            gamma = xi.sum(2).squeeze()

            # Updated gammas
            prod = (alpha[T - 1] * beta[T - 1]).unsqueeze(0)
            s = prod / prod.sum()
            updated_gamma = torch.cat([gamma, s], 0)
            if idx == 0:  # compute self.prob_state_1, only for future display
                self.prob_state_1.append(gamma[:, 0])

            # Store results for current sequence
            summed_xis.append(xi.sum(0))
            gammas.append(gamma)
            updated_gammas.append(updated_gamma)

        # Update A and pi according to formula 40a and 40b, and 109 for multiple observation sequences
        # Just sum up numerator and denominator and divide
        A_new = torch.stack(summed_xis).sum(0) / torch.stack(
            [gamma.sum(0).unsqueeze(1) for gamma in gammas]
        ).sum(0)

        pi_new = torch.stack([gamma[0, :] for gamma in gammas]).sum(0) / len(gammas)

        return pi_new, A_new, updated_gammas

    def re_estimate_emission(self, observed_sequences, updated_gammas):
        # Update B according to formula 40c
        states_marginal = torch.stack([gamma.sum(0) for gamma in updated_gammas]).sum(0)
        # One hot encoding buffer that you create out of the loop and just keep reusing
        emission_scores = []
        for observed_seq, gamma in zip(observed_sequences, updated_gammas):
            seq_one_hot = torch.zeros([len(observed_seq), self.M], dtype=torch.float64)
            seq_one_hot.scatter_(1, torch.tensor(observed_seq).unsqueeze(1), 1)
            local_emission_score = torch.matmul(seq_one_hot.transpose_(1, 0), gamma)
            emission_scores.append(local_emission_score)
        emission_score = torch.stack(emission_scores).sum(0)
        return emission_score / states_marginal

    def check_convergence(self, new_T0, new_transition, new_emission):
        delta_T0 = torch.max(torch.abs(self.pi - new_T0)).item() < self.epsilon
        delta_T = torch.max(torch.abs(self.A - new_transition)).item() < self.epsilon
        delta_E = torch.max(torch.abs(self.B - new_emission)).item() < self.epsilon

        return delta_T0 and delta_T and delta_E

    def expectation_maximization_step(self, obs_sequences, beta_normalization):
        # probability of emission sequence
        alphas, betas = [], []
        for obs_seq in obs_sequences:
            obs_prob_seq = self.B[obs_seq]
            alpha, beta = self.forward_backward(obs_prob_seq, beta_normalization)

            alphas.append(alpha)
            betas.append(beta)

        new_pi, new_A, updated_gammas = self.re_estimate_transition(obs_sequences, alphas, betas)

        new_B = self.re_estimate_emission(obs_sequences, updated_gammas)

        converged = self.check_convergence(new_pi, new_A, new_B)

        self.pi = new_pi
        self.B = new_B
        self.A = new_A

        return converged

    def Baum_Welch_EM(self, obs_sequences, beta_normalization="original"):
        converged = False
        for i in range(self.max_step):
            converged = self.expectation_maximization_step(obs_sequences, beta_normalization)
            display_progress("Baum-Welch optimization in progress", i + 1, self.max_step)
            if converged:
                print(f"\nConverged at step {i}")
                break

            print("\n------ Reconstructed model ------")
            print("Transition Matrix: ")
            print(self.A)
            print()
            print("Emission Matrix: ")
            print(self.B)
            print()
            print("Initial probabilities: ")
            print(self.pi)
            print()
            print("Reached Convergence: ")
            print(converged)

        # Return numpy arrays for easier display
        return (
            self.pi.detach().numpy(),
            self.A.detach().numpy(),
            self.B.detach().numpy(),
            converged,
        )



from torchvision.models import resnet50, ResNet50_Weights
import torch.nn as nn


class MetaphaseCnn(nn.Module):
    def __init__(self, nb_classes, nb_input_channels=3):
        super().__init__()

        # Define Conv2d if input does not have 3 channels
        self.input_channels = nb_input_channels
        # input_channels, output_channels, kernel_size
        self.conv = nn.Conv2d(nb_input_channels, 3, 3)

        weights = ResNet50_Weights.DEFAULT
        self.cnn = resnet50(weights=weights)
        num_features = self.cnn.fc.in_features
        self.cnn.fc = nn.Linear(num_features, nb_classes)

    def forward(self, x):
        if self.input_channels != 3:
            x = self.conv(x)
        return self.cnn(x)



import os
import pickle
from typing import Optional
import numpy as np

from ..factories.mt_cut_detection_factory import MtCutDetectionFactory
from ..models.tools import get_model_path
from ..utils.mitosis_track import MitosisTrack


def perform_mt_cut_detection(
    raw_video: np.ndarray,
    video_name: str,
    exported_mitoses_dir: str,
    scaler_path: Optional[str] = get_model_path("svc_scaler"),
    model_path: Optional[str] = get_model_path("svc_model"),
    hmm_bridges_parameters_file: Optional[str] = get_model_path("hmm_bridges_parameters"),
):
    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track = pickle.load(f)

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Generate a list of the first MT cut time for each mitosis track
    mt_cut_detector = MtCutDetectionFactory()
    for i, mitosis_track in enumerate(mitosis_tracks):
        print(f"Detect MT cut ({i+1}/{len(mitosis_tracks)})...")

        # Perform cut detection
        mt_cut_detector.update_mt_cut_detection(
            mitosis_track, raw_video, scaler_path, model_path, hmm_bridges_parameters_file
        )

        # Save updated mitosis track
        daughter_track_ids = ",".join([str(d) for d in mitosis_track.daughter_track_ids])
        state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
        save_path = os.path.join(
            exported_mitoses_dir,
            state_path,
        )
        with open(save_path, "wb") as f:
            pickle.dump(mitosis_track, f)



import os
from skimage import io

from cut_detector._widget import mid_body_detection
from cut_detector.data.tools import get_data_path


def test_open_mid_body_detection_widget():
    # Just try to open the widget
    mid_body_detection()


def test_mid_body_detection_widget(make_napari_viewer_proxy):
    viewer = make_napari_viewer_proxy()

    # Add video
    video = io.imread(os.path.join(get_data_path("videos"), "example_video.tif"))
    viewer.add_image(video, name="example_video")

    # Open the widget
    widget = mid_body_detection()

    # Run process
    widget(viewer.layers[0], get_data_path("mitoses"), get_data_path("tracks"), False, "")



import numpy as np
import napari


def main():
    # Create a Napari viewer
    viewer = napari.Viewer()

    # Create a blue rectangle
    rectangle_width = 50
    rectangle_height = 30
    rectangle_color = [0, 0, 1]  # Blue color in RGB

    # Define the rectangle vertices
    vertices = np.array(
        [
            [0, 0],
            [rectangle_width, 0],
            [rectangle_width, rectangle_height],
            [0, rectangle_height],
        ]
    )

    # Create a rectangle layer
    viewer.add_shapes(
        data=vertices,
        shape_type="rectangle",
        edge_color="transparent",
        face_color=rectangle_color,
    )

    # Display the Napari viewer
    napari.run()


if __name__ == "__main__":
    main()



import os
from typing import Optional
import numpy as np
from bigfish import stack, detection, plot
from skimage.morphology import extrema, opening
from scipy import ndimage
from scipy.optimize import linear_sum_assignment

from cnn_framework.utils.display_tools import display_progress

from ..constants.tracking import CYTOKINESIS_DURATION
from ..utils.mid_body_track import MidBodyTrack
from ..utils.image_tools import smart_cropping
from ..utils.mid_body_spot import MidBodySpot
from ..utils.mitosis_track import MitosisTrack
from ..utils.trackmate_track import TrackMateTrack


class MidBodyDetectionFactory:
    """
    Class to perform mid-body detection, tracking and filtering.

    Args:
        weight_mklp_intensity_factor (float): Weight of intensity in spot dist calculation
            (cf TrackMate).
        weight_sir_intensity_factor (float): Weight of sir intensity in spot distance calculation.
        mid_body_linking_max_distance (int): Maximum distance between two mid-bodies to link them.

        h_maxima_threshold (float): Threshold for h_maxima detection (default).

        sigma (float): Sigma for bigfish detection (unused).
        threshold (float): Threshold for bigfish detection (unused).

        cytokinesis_duration (int): Number of frames to look for mid-body in between cells.
        minimum_mid_body_track_length (int): Minimum spots in mid-body track to consider it.
    """

    def __init__(
        self,
        weight_mklp_intensity_factor=10.0,
        weight_sir_intensity_factor=3.33,
        mid_body_linking_max_distance=100,
        h_maxima_threshold=5.0,
        sigma=2.0,
        threshold=1.0,
        cytokinesis_duration=CYTOKINESIS_DURATION,
        minimum_mid_body_track_length=10,
    ) -> None:
        self.weight_mklp_intensity_factor = weight_mklp_intensity_factor
        self.weight_sir_intensity_factor = weight_sir_intensity_factor
        self.mid_body_linking_max_distance = mid_body_linking_max_distance
        self.h_maxima_threshold = h_maxima_threshold
        self.sigma = sigma
        self.threshold = threshold
        self.cytokinesis_duration = cytokinesis_duration
        self.minimum_mid_body_track_length = minimum_mid_body_track_length

    def update_mid_body_spots(
        self,
        mitosis_track: MitosisTrack,
        mitosis_movie: np.array,
        mask_movie: np.array,
        tracks: list[TrackMateTrack],
        path_output: Optional[str] = None,
    ) -> None:
        """
        Get spots of best mitosis track.

        Parameters
        ----------
        mitosis_movie: (T, H, W, C)
        mask_movie: (T, H, W)

        """

        spots_candidates = self._detect_mid_body_spots(
            mitosis_movie, mask_movie=mask_movie, path_output=path_output
        )
        mid_body_tracks = self._generate_tracks_from_spots(spots_candidates)
        kept_track = self._select_best_track(mitosis_track, mid_body_tracks, tracks, mitosis_movie)

        if kept_track is None:
            return

        # Keep only spots of best mitosis track
        for rel_frame, spot in kept_track.spots.items():
            frame = rel_frame + mitosis_track.min_frame
            mitosis_track.mid_body_spots[frame] = spot

    def _detect_mid_body_spots(
        self,
        mitosis_movie: np.array,
        mask_movie: Optional[np.array] = None,
        mid_body_channel=1,
        sir_channel=0,
        path_output: Optional[str] = None,
        mode="h_maxima",
    ) -> dict[int, list[MidBodySpot]]:
        """
        Parameters
        ----------
        mitosis_movie: (T, H, W, C)
        mask_movie: (T, H, W)

        Returns
        ----------
        spots_dictionary: dict[int, list[MidBodySpot]]
        """

        # Default mask is all ones
        if mask_movie is None:
            mask_movie = np.ones(mitosis_movie.shape[:-1])

        # Detect spots in each frame
        spots_dictionary = {}
        nb_frames = mitosis_movie.shape[0]
        for frame in range(nb_frames):
            display_progress(
                "Detect mid-body spots...",
                frame + 1,
                nb_frames,
                additional_message=f"Frame {frame + 1}/{nb_frames}",
            )

            mitosis_frame = mitosis_movie[frame, :, :, :].squeeze()  # H, W, C
            mask_frame = mask_movie[frame, :, :].squeeze()  # H, W
            spots = self.spot_detection(
                mitosis_frame,
                mask_frame,
                mid_body_channel,
                sir_channel,
                path_output=path_output,
                mode=mode,
                frame=frame,
            )

            # Update dictionary
            spots_dictionary[frame] = spots

        return spots_dictionary

    def spot_detection(
        self,
        image: np.array,
        mask: np.array,
        mid_body_channel: int,
        sir_channel: int,
        path_output: Optional[str] = None,
        mode="bigfish",
        frame=-1,
    ) -> list[MidBodySpot]:
        """
        Mode 'bigfish'
            threshold_1: sigma for log filter
            threshold_2: threshold for spots detection

        Mode 'h_maxima'
            threshold_1: threshold for h_maxima
            threshold_2: unused

        """

        image_sir = image[:, :, sir_channel]
        image_mklp = image[:, :, mid_body_channel]  #

        if mode == "bigfish":
            # Spots detection with bigfish functions
            filtered_image = stack.log_filter(image_mklp, sigma=self.sigma)
            filtered_image = image_mklp
            # Filter out spots which are not maximal or outside convex hull
            spots_mask = (filtered_image > 0) * mask
            # If mask is empty, skip frame
            if np.sum(spots_mask) == 0:
                spots = np.array([], dtype=np.int64).reshape((0, filtered_image.ndim))
            else:
                spots, _ = detection.spots_thresholding(
                    filtered_image, spots_mask.astype(np.bool), threshold=self.threshold
                )

        elif mode == "h_maxima":
            # Perform opening followed by closing to remove small spots
            filtered_image = opening(image_mklp, footprint=np.ones((3, 3)))
            # Get local maxima using h_maxima
            local_maxima = extrema.h_maxima(filtered_image, self.h_maxima_threshold)
            # Label spot regions
            labeled_local_maxima, nb_labels = ndimage.label(
                local_maxima, structure=np.ones((3, 3))
            )
            # Remove inconsistent labels
            # Threshold is computed as 99th percentile of image
            filtering_threshold = np.quantile(image_mklp.flatten(), 0.99)
            for label in range(1, nb_labels + 1):
                # Labels intensity in original image has to be higher than threshold
                if (
                    image_mklp[np.where(labeled_local_maxima == label)].mean()
                    < filtering_threshold
                ):
                    labeled_local_maxima[labeled_local_maxima == label] = 0
            # Re-label accordingly
            labeled_local_maxima, nb_labels = ndimage.label(
                labeled_local_maxima > 0, structure=np.ones((3, 3))
            )
            # Get center of mass to locate spots
            spots = ndimage.center_of_mass(
                local_maxima, labeled_local_maxima, range(1, nb_labels + 1)
            )
            spots = np.asarray(spots, dtype=np.int64)

            # Here, do something to retrieve mid_body area and/or circularity...

            if len(spots) == 0:
                spots = np.array([], dtype=np.int64).reshape((0, filtered_image.ndim))

        else:
            raise ValueError(f"Unknown mode: {mode}")

        if path_output is not None:
            # Check if directory exists
            if not os.path.exists(path_output):
                os.makedirs(path_output)
            # Count number of files in directory
            nb_files = len(os.listdir(path_output))
            plot.plot_detection(
                filtered_image,
                spots,
                contrast=True,
                path_output=os.path.join(path_output, f"spot_detection_{nb_files+1}.png"),
                show=False,
            )

        # Convert spots to MidBodySpot objects (switch (y, x) to (x, y))
        mid_body_spots = [
            MidBodySpot(
                frame,
                (position[1], position[0]),
                intensity=self._get_average_intensity(position, image_mklp),
                sir_intensity=self._get_average_intensity(position, image_sir),
            )
            for position in spots
        ]

        return mid_body_spots

    @staticmethod
    def _get_average_intensity(position: tuple[int], image: np.array, margin=1) -> int:
        """
        Parameters
        ----------
        position: (y, x)
        image: (H, W)
        margin: int

        Returns
        ----------
        average_intensity: int
        """
        # Get associated crop
        crop = smart_cropping(
            image, margin, position[1], position[0], position[1] + 1, position[0] + 1
        )

        # Return average intensity
        return int(np.mean(crop))

    def _update_spots_hereditary(
        self, spots1: list[MidBodySpot], spots2: list[MidBodySpot]
    ) -> None:
        """
        Link spots together using Hungarian algorithm.
        """
        # Ignore empty spots list
        if len(spots1) == 0 or len(spots2) == 0:
            return

        # Create cost matrix
        # https://imagej.net/plugins/trackmate/algorithms
        cost_matrix = np.zeros((len(spots1) + len(spots2), len(spots1) + len(spots2)))
        max_cost = 0
        for i, spot1 in enumerate(spots1):
            for j, spot2 in enumerate(spots2):
                intensity_penalty = (
                    3
                    * self.weight_mklp_intensity_factor
                    * np.abs(spot1.intensity - spot2.intensity)
                    / (spot1.intensity + spot2.intensity)
                )
                sir_intensity_penalty = (
                    3
                    * self.weight_sir_intensity_factor
                    * np.abs(spot1.sir_intensity - spot2.sir_intensity)
                    / (spot1.sir_intensity + spot2.sir_intensity)
                )
                penalty = 1 + intensity_penalty + sir_intensity_penalty
                distance = np.linalg.norm(np.array(spot1.position) - np.array(spot2.position))
                if distance > self.mid_body_linking_max_distance:
                    cost_matrix[i, j] = np.inf
                else:
                    # Compared to original TrackMate algorithm, remove square to penalize no attribution to the closest spot
                    cost_matrix[i, j] = (penalty * distance) ** 1
                    max_cost = max(max_cost, cost_matrix[i, j])

        min_cost = 0 if np.max(cost_matrix) == 0 else np.min(cost_matrix[np.nonzero(cost_matrix)])

        cost_matrix[len(spots1) :, : len(spots2)] = max_cost * 1.05  # bottom left
        cost_matrix[: len(spots1), len(spots2) :] = max_cost * 1.05  # top right
        cost_matrix[len(spots1) :, len(spots2) :] = min_cost  # bottom right

        # Hungarian algorithm
        row_ind, col_ind = linear_sum_assignment(cost_matrix)

        # Update parent and child spots
        for i, j in zip(row_ind, col_ind):
            if i < len(spots1) and j < len(spots2):
                spots1[i].child_spot = spots2[j]
                spots2[j].parent_spot = spots1[i]

    def _generate_tracks_from_spots(
        self, spots_candidates: dict[int, list[MidBodySpot]]
    ) -> list[MidBodyTrack]:
        """
        Use spots linked together to generate tracks.

        Parameters
        ----------
        spots_candidates : {frame: [MidBodySpot]}

        Returns
        ----------
        tracks: [MidBodyTrack]

        """

        # Update parent and child spots
        for frame in spots_candidates.keys():
            # Ignore if no spot detected in next frame
            if not frame + 1 in spots_candidates:
                continue
            self._update_spots_hereditary(spots_candidates[frame], spots_candidates[frame + 1])

        tracks = []
        for spots in spots_candidates.values():
            for spot in spots:
                if spot.track_id is None:
                    # Create new track
                    track_id = len(tracks)
                    new_track = MidBodyTrack(track_id)
                    new_track.add_spot(spot)
                    tracks.append(new_track)
        return tracks

    def _select_best_track(
        self,
        mitosis_track: MitosisTrack,
        mid_body_tracks: list[MidBodyTrack],
        trackmate_tracks: list[TrackMateTrack],
        mitosis_movie: np.array,
        sir_channel=0,
    ) -> MidBodyTrack:
        """
        Select best track from mid-body tracks.
        """
        mother_track, daughter_tracks = mitosis_track.get_mother_daughters_tracks(trackmate_tracks)
        # NB: only first daughter is considered
        daughter_track = daughter_tracks[0]

        expected_positions = {}
        for frame in range(daughter_track.start, daughter_track.start + self.cytokinesis_duration):
            # If one cell does not exist anymore, stop
            if frame not in daughter_track.track_spots or frame not in mother_track.track_spots:
                continue
            # Compute mid-body expected relative position at current frame
            closest_points = []
            min_distance = np.inf
            for mother_point in mother_track.track_spots[frame].spot_points:
                position_mother = [
                    int(mother_point[0]) - mitosis_track.position.min_x,
                    int(mother_point[1]) - mitosis_track.position.min_y,
                ]
                for daughter_point in daughter_track.track_spots[frame].spot_points:
                    position_daughter = [
                        int(daughter_point[0]) - mitosis_track.position.min_x,
                        int(daughter_point[1]) - mitosis_track.position.min_y,
                    ]
                    distance = np.linalg.norm(
                        [a - b for a, b in zip(position_mother, position_daughter)]
                    )
                    if distance < min_distance:
                        min_distance = distance
                        closest_points = [(position_mother, position_daughter)]
                    if distance == min_distance:
                        closest_points.append((position_mother, position_daughter))

            mid_body_position = np.mean(closest_points, axis=0)
            mid_body_position = np.mean(mid_body_position, axis=0)
            expected_positions[frame - mitosis_track.min_frame] = mid_body_position

        # Remove wrong tracks by keeping only tracks with at least minimum_track_length points
        mid_body_tracks = [
            track for track in mid_body_tracks if track.length > self.minimum_mid_body_track_length
        ]

        # Compute mean intensity on sir-tubulin channel for each track
        image_sir = mitosis_movie[:, :, :, sir_channel]
        sir_intensity_track = [0 for _ in mid_body_tracks]
        for idx, track in enumerate(mid_body_tracks):
            abs_track_frames = [
                frame + mitosis_track.min_frame for frame in list(track.spots.keys())
            ]
            abs_min_frame = mitosis_track.key_events_frame["cytokinesis"]  # Cytokinesis start
            abs_max_frame = abs_min_frame + int(self.cytokinesis_duration / 2)
            if abs_min_frame > abs_track_frames[-1] or abs_max_frame < abs_track_frames[0]:
                sir_intensity_track[idx] = -np.inf
            frame_count = 0
            for frame in range(abs_min_frame, abs_max_frame + 1):
                if frame not in abs_track_frames:
                    continue
                frame_count += 1
                track_position = track.spots[frame - mitosis_track.min_frame].position
                sir_intensity_track[idx] += image_sir[
                    frame - mitosis_track.min_frame, track_position[1], track_position[0]
                ]

            if frame_count < (abs_max_frame - abs_min_frame + 1) / 2:
                sir_intensity_track[idx] = -np.inf
            else:
                sir_intensity_track[idx] /= frame_count

        # Get list of expected distances
        expected_distances = []
        for track in mid_body_tracks:
            val = track.get_expected_distance(
                expected_positions, self.mid_body_linking_max_distance
            )
            expected_distances.append(val)

        # Assert lists have same length for next function
        assert len(expected_distances) == len(mid_body_tracks)
        assert len(sir_intensity_track) == len(mid_body_tracks)

        # function to sort tracks by expected distance and intensity
        def func_sir_intensity(track):
            a = expected_distances[mid_body_tracks.index(track)]
            b = sir_intensity_track[mid_body_tracks.index(track)]
            return a - 0.5 * b

        # Remove tracks with infinite func value
        fun_values = []
        final_tracks = []
        for track in mid_body_tracks:
            fun_values.append(func_sir_intensity(track))
            if func_sir_intensity(track) != np.inf:
                final_tracks.append(track)

        # Sort tracks by func value
        sorted_tracks = sorted(final_tracks, key=func_sir_intensity)
        return sorted_tracks[0] if len(sorted_tracks) > 0 else None



from __future__ import annotations
from typing import Optional

import numpy as np
from shapely.geometry.polygon import Polygon
from shapely import distance


class TrackMateSpot:
    """
    Parse TrackMate spot from xml file.
    """

    def __init__(self, trackmate_spot, raw_video_shape: list[int]):
        self.x = float(trackmate_spot["@POSITION_X"])
        self.y = float(trackmate_spot["@POSITION_Y"])
        self.frame = int(trackmate_spot["@FRAME"])
        self.id = int(trackmate_spot["@ID"])
        self.track_id: int = -1

        # Get min and max positions
        raw_positions: str = trackmate_spot["#text"].split(" ")
        rel_positions_x = [float(raw_positions[i]) for i in range(0, len(raw_positions), 2)]
        rel_positions_y = [float(raw_positions[i]) for i in range(1, len(raw_positions), 2)]

        self.rel_min_x, self.rel_max_x = min(rel_positions_x), max(rel_positions_x)
        self.rel_min_y, self.rel_max_y = min(rel_positions_y), max(rel_positions_y)

        # Get spot points
        positions_x = [
            int(self.x + float(raw_positions[i])) for i in range(0, len(raw_positions), 2)
        ]
        positions_y = [
            int(self.y + float(raw_positions[i])) for i in range(1, len(raw_positions), 2)
        ]
        self.spot_points = [[x, y] for x, y in zip(positions_x, positions_y)]

        # Get min and max positions
        self.abs_min_x, self.abs_max_x = (
            int(self.x + self.rel_min_x),
            int(self.x + self.rel_max_x),
        )
        self.abs_min_y, self.abs_max_y = (
            int(self.y + self.rel_min_y),
            int(self.y + self.rel_max_y),
        )

        # Clip to video size
        self.abs_min_x, self.abs_max_x = (
            max(self.abs_min_x, 0),
            min(self.abs_max_x, raw_video_shape[2]),
        )
        self.abs_min_y, self.abs_max_y = (
            max(self.abs_min_y, 0),
            min(self.abs_max_y, raw_video_shape[1]),
        )

        # Phase predicted by model
        self.predicted_phase: Optional[int] = None

        # Corresponding (closest) metaphase spot in track
        self.corresponding_metaphase_spot = None

    def distance_to(self, other_spot: TrackMateSpot):
        return np.sqrt((self.x - other_spot.x) ** 2 + (self.y - other_spot.y) ** 2)

    def is_stuck_to(self, other_spot: TrackMateSpot, maximum_stuck_distance: float):
        """
        Distance between two spots hulls.
        """
        return (
            distance(Polygon(self.spot_points), Polygon(other_spot.spot_points))
            < maximum_stuck_distance
        )



import os
import urllib.request

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def get_data_path(data_name: str) -> None:
    """
    Returns absolute path to model file.
    """
    sub_folder_to_create = data_name
    if data_name == "mitoses":
        files = ["example_video_mitosis_0_4_to_0.bin"]
    elif data_name == "models":
        files = ["example_video_model.xml"]
    elif data_name == "tracks":
        sub_folder_to_create = f"{data_name}/example_video"
        files = [
            "example_video/track_0.bin",
            "example_video/track_1.bin",
            "example_video/track_2.bin",
            "example_video/track_3.bin",
            "example_video/track_4.bin",
        ]
    elif data_name == "results":
        files = []  # no files to download here
    elif data_name == "videos":
        files = ["example_video.tif"]
    else:
        raise ValueError(f"Unknown data name: {data_name}")

    local_path = os.path.join(CURRENT_DIR, data_name)

    if sub_folder_to_create is not None:
        os.makedirs(os.path.join(CURRENT_DIR, sub_folder_to_create), exist_ok=True)

    for file in files:
        file_local_path = os.path.join(local_path, file)
        if not os.path.exists(file_local_path):
            print(f"Downloading data {data_name}...")
            urllib.request.urlretrieve(
                f"https://raw.githubusercontent.com//15bonte/cut-detector-models/main/data/{data_name}/{file}",
                file_local_path,
            )

    return local_path



import os
import sys
from typing import Optional
import torch
import imagej
import scyjava as sj
from cellpose import models

from ..models.tools import get_model_path
from ..constants.tracking import (
    AUGMENT,
    CELLPROB_THRESHOLD,
    FLOW_THRESHOLD,
    GAP_CLOSING_MAX_DISTANCE_RATIO,
    LINKING_MAX_DISTANCE_RATIO,
    MAX_FRAME_GAP,
)


def perform_tracking(
    video_path: str,
    fiji_path: str,
    save_folder: str,
    model_path: Optional[str],
    fast_mode: bool,
) -> None:
    # Load default model if necessary
    if model_path is None:
        model_path = get_model_path("segmentation_model")

    # Create save directory if it doesn't exist
    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    ij_instance = imagej.init(fiji_path, mode="interactive")

    bf_plugin = sj.jimport("loci.plugins.BF")
    importer_options = sj.jimport("loci.plugins.in.ImporterOptions")

    # Import TrackMate via scyjava
    track_mate_model = sj.jimport(
        "fiji.plugin.trackmate.Model"
    )  # class in charge of storing the data
    track_mate_settings = sj.jimport(
        "fiji.plugin.trackmate.Settings"
    )  # class storing the fields that will configure TrackMate and pilot how the data is created
    track_mate = sj.jimport("fiji.plugin.trackmate.TrackMate")
    logger = sj.jimport("fiji.plugin.trackmate.Logger")
    lap_utils = sj.jimport("fiji.plugin.trackmate.tracking.jaqaman.LAPUtils")
    sparse_lap_tracker_factory = sj.jimport(
        "fiji.plugin.trackmate.tracking.jaqaman.SparseLAPTrackerFactory"
    )
    file = sj.jimport("java.io.File")
    tm_xml_writer = sj.jimport("fiji.plugin.trackmate.io.TmXmlWriter")
    tracker_keys = sj.jimport("fiji.plugin.trackmate.tracking.TrackerKeys")
    track_mate_track_analyzer_provider = sj.jimport(
        "fiji.plugin.trackmate.providers.TrackAnalyzerProvider"
    )
    track_mate_edge_analyzer_provider = sj.jimport(
        "fiji.plugin.trackmate.providers.EdgeAnalyzerProvider"
    )

    if fast_mode:
        try:
            cellpose_detector_factory = sj.jimport(
                "fiji.plugin.trackmate.cellpose.tbonte.CellposeDetectorFactory"
            )
            pretrained_model = sj.jimport(
                "fiji.plugin.trackmate.cellpose.tbonte.CellposeSettings.PretrainedModel"
            )
        except TypeError:
            print("No Trackmate plugin found for fast mode. Using usual segmentation instead.")
            fast_mode = False

    if not fast_mode:
        cellpose_detector_factory = sj.jimport(
            "fiji.plugin.trackmate.cellpose.CellposeDetectorFactory"
        )
        pretrained_model = sj.jimport(
            "fiji.plugin.trackmate.cellpose.CellposeSettings.PretrainedModel"
        )

    # Skip if file already exists
    video_file_name = os.path.basename(video_path).split(".")[0]
    out_file = os.path.join(save_folder, f"{video_file_name}_model.xml")
    if os.path.exists(out_file):
        print(f"File {out_file} already exists. Skipping.")
        return

    # Get currently selected image
    options = importer_options()
    options.setColorMode(importer_options.COLOR_MODE_GRAYSCALE)
    options.setId(video_path)
    imps = bf_plugin.openImagePlus(options)
    imp = imps[0]

    # Swap Z and T dimensions if necessary
    dims = imp.getDimensions()
    if imp.dims[-1] == "Z":
        imp.setDimensions(sj.to_java(dims[2]), sj.to_java(dims[4]), sj.to_java(dims[3]))
        print("Swapping Z and T dimensions")

    # Get the average spot size to define max linking distances
    average_spot_size = models.CellposeModel(pretrained_model=[model_path]).diam_labels

    # Create model object
    model = track_mate_model()

    # Send all messages to ImageJ log window.
    model.setLogger(logger.IJ_LOGGER)

    # Prepare settings object
    settings = track_mate_settings(imp)

    # Configure detector - We use the Strings for the keys
    settings.detectorFactory = cellpose_detector_factory()
    settings.detectorSettings["TARGET_CHANNEL"] = sj.to_java(3)
    settings.detectorSettings["OPTIONAL_CHANNEL_2"] = sj.to_java(0)
    settings.detectorSettings["CELLPOSE_PYTHON_FILEPATH"] = sys.executable
    settings.detectorSettings["CELLPOSE_MODEL_FILEPATH"] = model_path
    settings.detectorSettings["CELLPOSE_MODEL"] = pretrained_model.CUSTOM
    settings.detectorSettings["CELL_DIAMETER"] = sj.to_java(0.0, type="double")
    settings.detectorSettings["USE_GPU"] = True if torch.cuda.is_available() else False
    settings.detectorSettings["SIMPLIFY_CONTOURS"] = True

    if fast_mode:
        settings.detectorSettings["FLOW_THRESHOLD"] = FLOW_THRESHOLD
        settings.detectorSettings["CELLPROB_THRESHOLD"] = CELLPROB_THRESHOLD
        settings.detectorSettings["AUGMENT"] = AUGMENT

    # Configure tracker
    settings.trackerFactory = sparse_lap_tracker_factory()
    settings.trackerSettings = lap_utils.getDefaultSegmentSettingsMap()  # almost good enough
    settings.trackerSettings["LINKING_MAX_DISTANCE"] = (
        LINKING_MAX_DISTANCE_RATIO * average_spot_size
    )
    settings.trackerSettings["ALLOW_GAP_CLOSING"] = True
    settings.trackerSettings["GAP_CLOSING_MAX_DISTANCE"] = (
        GAP_CLOSING_MAX_DISTANCE_RATIO * average_spot_size
    )
    settings.trackerSettings["MAX_FRAME_GAP"] = sj.to_java(MAX_FRAME_GAP)
    settings.trackerSettings["ALLOW_TRACK_MERGING"] = False
    settings.trackerSettings["ALLOW_TRACK_SPLITTING"] = False
    settings.trackerSettings[
        "LINKING_FEATURE_PENALTIES"
    ] = tracker_keys.DEFAULT_LINKING_FEATURE_PENALTIES

    settings.initialSpotFilterValue = -1.0

    # Add useful track analyzers
    track_analyzer_provider = track_mate_track_analyzer_provider()
    settings.addTrackAnalyzer(track_analyzer_provider.getFactory("Track duration"))
    settings.addTrackAnalyzer(track_analyzer_provider.getFactory("Track index"))

    edge_analyzer_provider = track_mate_edge_analyzer_provider()
    settings.addEdgeAnalyzer(edge_analyzer_provider.getFactory("Edge target"))

    # Instantiate plugin
    trackmate = track_mate(model, settings)

    # Process
    process_ok = trackmate.checkInput()
    if not process_ok:
        sys.exit(str(trackmate.getErrorMessage()))

    process_ok = trackmate.process()
    if not process_ok:
        sys.exit(str(trackmate.getErrorMessage()))

    # Echo results with the logger we set at start:
    model.getLogger().log(str(model))

    out_file_model = file(save_folder, f"{video_file_name}_model.xml")
    writer = tm_xml_writer(out_file_model)
    writer.appendModel(model)
    writer.appendSettings(settings)
    writer.writeToFile()

    # Force exit
    ij_instance.dispose()



from cnn_framework.utils.model_params.base_model_params import BaseModelParams
from cnn_framework.utils.dimensions import Dimensions


class MetaphaseCnnModelParams(BaseModelParams):
    """
    Metaphase CNN model params.
    """

    def __init__(self):
        super().__init__("metaphase_cnn")

        self.input_dimensions = Dimensions(height=256, width=256)

        self.num_epochs = 50
        self.learning_rate = 0.1

        self.nb_classes = 3
        self.class_names = ["Interphase", "Metaphase", "Death"]

        self.c_indexes = [0, 1, 2]
        self.z_indexes = [0]
